import matplotlib.pyplot as plt
import numpy as np
import scanpy as sc
import scvelo as scv
import pandas as pd
import warnings
warnings.filterwarnings("ignore")


def get_start_node(adata, select_info, group_key, njobs=1, mode="stochastic"):
    '''
    :param adata: anndata to use
    :param select_info: dataframe information for selected clusters and edges
    :param group_key: cell cluster key to use
    :param njobs: number of parallel jobs
    :param mode: start node finding mode
    :return: string for start node cluster
    '''
    # run scvelo and get root/end score
    if "velocity_graph" not in adata.uns.keys():
        if mode == "stochastic":
            scv.tl.velocity(adata, mode='stochastic')
        elif mode == "dynamical":
            scv.tl.recover_dynamics(adata, n_jobs=njobs)
            scv.tl.velocity(adata, mode="dynamical")
        else:
            raise KeyError("mode must be stochastic or dynamical")

    scv.tl.velocity_graph(adata, n_jobs=njobs)
    scv.tl.terminal_states(adata)

    select_clusters = select_info.index.values
    select_index = adata.obs[group_key].isin(select_clusters)
    adata = adata[select_index, :]

    root_score = np.array(adata.obs.groupby(group_key)["root_cells"].mean())
    end_score = np.array(adata.obs.groupby(group_key)["end_points"].mean())

    root_order = abs(np.sort(-root_score))
    end_order = abs(np.sort(-end_score))

    if np.max(root_score) > 0.1:
        flag = 0
        root_index = np.where(root_score == root_order[flag])[0][0]
        while flag < len(root_score) - 1:
            if end_score[root_index] > root_score[root_index] or root_score[root_index] <= 0.1:
                flag += 1
                root_index = np.where(root_score == root_order[flag])[0][0]
            else:
                break

        if end_score[root_index] <= root_score[root_index] and root_score[root_index] > 0.1:
            start_node = select_info.index[root_index]
            return start_node

        elif np.max(end_score) < 0.1:
            root_index = np.where(root_score == root_order[0])[0][0]
            start_node = select_info.index[root_index]
            return start_node

        else:
            flag = 0
            end_index = np.where(end_score == end_order[flag])[0][0]
            while flag < len(end_score) - 1:
                if root_score[end_index] > end_score[end_index] or end_score[end_index] <= 0.1:
                    flag += 1
                    end_index = np.where(end_score == end_order[flag])[0][0]
                else:
                    break
            if root_score[end_index] <= end_score[end_index] and end_score[end_index] > 0.1:
                start_node = select_info.index[end_index]
                return start_node
            else:
                root_index = np.where(root_score == root_order[0])[0][0]
                start_node = select_info.index[root_index]
                return start_node
    else:
        flag = 0
        end_index = np.where(end_score == end_order[flag])[0][0]
        while flag < len(end_score) - 1:
            if root_score[end_index] > end_score[end_index] or end_score[end_index] <= 0.1:
                flag += 1
                end_index = np.where(end_score == end_order[flag])[0][0]
            else:
                break
        if root_score[end_index] <= end_score[end_index] and end_score[end_index] > 0.1:
            start_node = select_info.index[end_index]
            return start_node
        else:
            root_index = np.where(root_score == root_order[0])[0][0]
            start_node = select_info.index[root_index]
            return start_node


if __name__ == '__main__':
    # test
    frame_path = "D:/cuhk/project/velocity/dataset/scRNA-seq/data_frame.csv"
    data_frame = pd.read_csv(frame_path, index_col=0)

    for data_name in data_frame.index:
        data_path = data_frame.loc[data_name]["path"]
        group_key = data_frame.loc[data_name]["group_key"]
        adata = sc.read_h5ad(data_path)

        scv.tl.velocity(adata, mode='stochastic')
        scv.tl.velocity_graph(adata, n_jobs=20)

        scv.tl.terminal_states(adata)
        ax = scv.pl.scatter(adata, color=['root_cells', 'end_points'], show=False)
        plt.show()

        root_score = adata.obs.groupby(group_key)["root_cells"].mean()
        print("\n", data_name, "\nroot score:\n", root_score)

        terminal_score = adata.obs.groupby(group_key)["end_points"].mean()
        print("\nterminal score:\n", terminal_score)









import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from numba import jit
from warnings import simplefilter
simplefilter(action='ignore', category=FutureWarning)


@jit(['Tuple((f8[:,:], f8[:,:]))(f8[:,:], i4, f8)'], nopython=True, fastmath=True, cache=True)
def bin_split_dataset(dataSet, feature, value):
    mat0 = dataSet[np.where(dataSet[:, feature] > value)[0], :]
    mat1 = dataSet[np.where(dataSet[:, feature] <= value)[0], :]
    return mat0, mat1


@jit(nopython=True, fastmath=True, cache=True)
def linear_solve(dataSet):
    m, n = np.shape(dataSet)
    X = np.ones((m, n))
    X[:, 1:n] = dataSet[:, 0:n-1]
    Y = dataSet[:, -1]
    xTx = np.dot(X.T, X)
    if np.linalg.det(xTx) == 0.0:
        raise NameError('This matrix is singular, cannot do inverse, try increasing dur')
    ws = np.dot(np.linalg.inv(xTx), np.dot(X.T, np.copy(Y)))
    return ws, X, Y


@jit(['f8[:](f8[:,:])'], nopython=True, fastmath=True, cache=True)
def model_leaf(dataSet):
    ws, X, Y = linear_solve(dataSet)
    return ws


@jit(['f8(f8[:,:])'], nopython=True, fastmath=True, cache=True)
def model_error(dataSet):
    ws, X, Y = linear_solve(dataSet)
    yHat = np.dot(np.copy(X), np.copy(ws))
    return np.sum(np.power(Y - yHat, 2))


@jit(['Tuple((i4, f8, f8[:], f8, f8))(f8[:,:], i4, i4)'], nopython=True, fastmath=True, cache=True)
def choose_best_split(dataSet, dur, step):
    m, n = np.shape(dataSet)
    S = model_error(dataSet)
    bestS = np.inf
    bestIndex = 0
    bestValue = 0
    for featIndex in range(n-1):
        l, u = int(np.min(dataSet[:, featIndex])), int(np.max(dataSet[:, featIndex]))
        for splitVal in range(l, u+1, step):
            mat0, mat1 = bin_split_dataset(dataSet, featIndex, splitVal)
            if (np.shape(mat0)[0] < dur) or (np.shape(mat1)[0] < dur):
                continue
            newS = model_error(mat0) + model_error(mat1)
            if newS < bestS:
                bestIndex = featIndex
                bestValue = splitVal
                bestS = newS

    return bestIndex, bestValue, model_leaf(dataSet), S, bestS


def is_tree(obj):
    return isinstance(obj, dict)


def model_tree_eval(model, inDat):
    n = np.shape(inDat)[1]
    X = np.ones((1, n+1))
    X[:, 1:n+1] = inDat
    return float(np.dot(X, model))


def tree_forecast(tree, inData):
    if not is_tree(tree):
        return model_tree_eval(tree, inData)
    if inData[tree['spInd']] > tree['spVal']:
        if is_tree(tree['left']):
            return tree_forecast(tree['left'], inData)
        else:
            return model_tree_eval(tree['left'], inData)
    else:
        if is_tree(tree['right']):
            return tree_forecast(tree['right'], inData)
        else:
            return model_tree_eval(tree['right'], inData)


def create_forecast(tree, testData):
    m = len(testData)
    yHat = np.zeros(m)
    for i in range(m):
        yHat[i] = tree_forecast(tree, np.mat(testData[i]))
    return yHat


def draw(dataSet, tree):
    plt.scatter(dataSet[:, 0], dataSet[:, 1], s=5)
    yHat = create_forecast(tree, dataSet[:, 0])
    plt.plot(dataSet[:, 0], yHat, linewidth=2.0, color='red')
    plt.show()


def create_tree(dataSet, rate, dur, step):
    if len(set(dataSet[:, -1])) == 1:
        return model_leaf(dataSet)
    feat, val, ws, S, bestS = choose_best_split(dataSet, dur, step)
    if (S - bestS) < rate:
        return ws
    retTree = {}
    retTree['spInd'] = feat
    retTree['spVal'] = val
    lSet, rSet = bin_split_dataset(dataSet, feat, val)
    retTree['left'] = create_tree(lSet, rate, dur, step)
    retTree['right'] = create_tree(rSet, rate, dur, step)
    return retTree


if __name__ == '__main__':
    import tushare as ts

    df = ts.get_k_data(code='002230', start='2000-01-01')
    e = pd.DataFrame()
    e['idx'] = df.index
    e['close'] = df['close']
    arr = np.array(e)

    tree = create_tree(arr, 100, 10, 1)
    draw(arr, tree)





import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.utils import data
import os
from tqdm import tqdm
import numpy as np
from ..utils.velocity_genes import compute_velocity_genes


class DNNLayer(nn.Module):
    def __init__(self, n_genes, n_dims=[256, 64]):
        super(DNNLayer, self).__init__()
        self.n_genes = n_genes

        self.alpha, self.beta, self.gamma = None, None, None
        
        self.layers = nn.ModuleList()
        # input layer
        self.layers.append(nn.Sequential(nn.Linear(self.n_genes * 2, n_dims[0]), nn.ReLU(True)))
        # hidden layers
        for i in range(len(n_dims) - 1):
            self.layers.append(nn.Sequential(nn.Linear(n_dims[i], n_dims[i + 1]), nn.ReLU(True)))

        out_layer_dim = n_genes * 3
        self.layers.append(nn.Sequential(nn.Linear(n_dims[-1], out_layer_dim), nn.ReLU(True)))

    def forward(self, x_u, x_s):
        h = torch.cat([x_u, x_s], dim=1)
        for i, layer in enumerate(self.layers):
            h = layer(h)
        x = h

        self.alpha = x[:, 0: self.n_genes]
        self.beta = x[:, self.n_genes: 2 * self.n_genes]
        self.gamma = x[:, 2 * self.n_genes: 3 * self.n_genes]

        v_u = self.alpha - self.beta * x_u
        v_s = self.beta * x_u - self.gamma * x_s

        return v_u, v_s

    def get_current_batch_kinetic_rates(self):
        return self.current_kinetic_rates


class DNN(object):
    def __init__(self, x_u, x_s, d_nn, knn, n_dims, normalize=False):

        assert x_u.shape[0] == x_s.shape[0] and x_u.shape[1] == x_s.shape[1],\
            "Data dimensions are not consistent"
        # configurations about inputs u and s
        self.x_u = torch.from_numpy(x_u)
        self.x_s = torch.from_numpy(x_s)

        self.x_u_norm = self.x_u / torch.nan_to_num(torch.std(self.x_u, dim=0, keepdim=True), 1)
        self.x_s_norm = self.x_s / torch.nan_to_num(torch.std(self.x_s, dim=0, keepdim=True), 1)

        self.d_nn = torch.from_numpy(d_nn).to(torch.float32)
        self.knn = torch.from_numpy(knn)

        self.n_cells, self.n_genes = x_u.shape[0], x_s.shape[1]
        self.n_dims = n_dims

        # initialize DNN
        self.dnn = DNNLayer(self.n_genes, n_dims)
        self.normalize = normalize

        # d_u, d_s
        n_neighs_d = torch.sum(self.d_nn, dim=1)
        n_neighs_d[torch.where(n_neighs_d == 0)] = 1
        self.d_u = torch.mm(self.d_nn, self.x_u) / n_neighs_d.unsqueeze(1)
        self.d_s = torch.mm(self.d_nn, self.x_s) / n_neighs_d.unsqueeze(1)
        self.v_u, self.v_s = self.d_u - self.x_u, self.d_s - self.x_s

    def pretrain(self, n_epochs=100, lr=1e-3, batch_size=128, loss_fun="mse", only_s=False):
        mse = nn.MSELoss()
        cos_sim = nn.CosineSimilarity(dim=1)
        parameter = self.dnn.parameters()
        optimizer = optim.Adam(parameter, amsgrad=True, lr=lr)
        batch_size = self.n_cells if batch_size is None else batch_size
        self.dnn.train()

        if self.normalize:
            train_data = data.TensorDataset(self.x_u_norm, self.x_s_norm, torch.arange(self.n_cells))
            train_loader = data.DataLoader(train_data, batch_size=batch_size, shuffle=False)
        else:
            train_data = data.TensorDataset(self.x_u, self.x_s, torch.arange(self.n_cells))
            train_loader = data.DataLoader(train_data, batch_size=batch_size, shuffle=False)

        pbar = tqdm(range(n_epochs))
        for epoch in pbar:
            loss_var = []
            for u, s, idx in train_loader:
                optimizer.zero_grad()
                
                v_u, v_s = self.dnn(u, s)

                if loss_fun == "mse":
                    if only_s:
                        loss = mse(v_s, self.v_s[idx, :])
                    else:
                        loss = mse(v_u, self.v_u[idx, :]) + mse(v_s, self.v_s[idx, :])
                elif loss_fun == "cos":
                    if only_s:
                        loss = -torch.mean(cos_sim(v_s, self.v_s[idx, :]))
                    else:
                        loss = -torch.mean(cos_sim(v_u, self.v_u[idx, :])) - torch.mean(cos_sim(v_s, self.v_s[idx, :]))
                else:
                    raise KeyError("No such loss function for pretraining")

                loss.backward()
                optimizer.step()
                loss_var.append(loss.clone().detach().numpy())

            pbar.set_description("Pretraining")
            if loss_fun == "mse":
                pbar.set_postfix(loss=np.around(np.mean(loss_var), 3))
            elif loss_fun == "cos":
                pbar.set_postfix(loss=np.around(-np.mean(loss_var), 3))

    def fit(self, n_epochs=100, lr=1e-3, batch_size=64, constrain=True, loss_fun="mse", alpha_1=1, alpha_2=1,
            only_s=False, same_c=None):
        mse = nn.MSELoss()
        cos_sim = nn.CosineSimilarity(dim=1)
        same_c = torch.from_numpy(same_c).to(torch.float32) if same_c is not None else None
        parameter = self.dnn.parameters()
        batch_size = self.n_cells if batch_size is None else batch_size
        optimizer = optim.Adam(parameter, amsgrad=True, lr=lr)
        self.dnn.train()

        if self.normalize:
            train_data = data.TensorDataset(self.x_u_norm, self.x_s_norm, self.d_u, self.d_s, torch.arange(self.n_cells))
            train_loader = data.DataLoader(train_data, batch_size=batch_size, shuffle=False)
        else:
            train_data = data.TensorDataset(self.x_u, self.x_s, self.d_u, self.d_s, torch.arange(self.n_cells))
            train_loader = data.DataLoader(train_data, batch_size=batch_size, shuffle=False)

        pbar = tqdm(range(n_epochs))
        for epoch in pbar:
            mse_u_, mse_s_, cos_u_, cos_s_, cos_xu_, cos_xs_ = [], [], [], [], [], []

            for u, s, d_u, d_s, idx in train_loader:
                optimizer.zero_grad()
                v_u, v_s = self.dnn(u, s)

                # cosine similarity for velocity
                norm_v_u = torch.norm(v_u, p=2, dim=1, keepdim=True)
                norm_v_s = torch.norm(v_s, p=2, dim=1, keepdim=True)
                cos_u = torch.mm(v_u, v_u.t()) / (torch.mm(norm_v_u, norm_v_u.t()) + 1e-6)
                cos_s = torch.mm(v_s, v_s.t()) / (torch.mm(norm_v_s, norm_v_s.t()) + 1e-6)

                if same_c is None:
                    n_neighs = torch.mean(torch.sum(self.knn[idx, :][:, idx], dim=1))
                    cos_u = torch.sum(torch.mean(cos_u * self.knn[idx, :][:, idx], dim=1)) / (n_neighs + 1e-6)
                    cos_s = torch.sum(torch.mean(cos_s * self.knn[idx, :][:, idx], dim=1)) / (n_neighs + 1e-6)
                else:
                    n_neighs = torch.mean(torch.sum(same_c * self.knn[idx, :][:, idx], dim=1))
                    cos_u = torch.sum(torch.mean(cos_u * same_c * self.knn[idx, :][:, idx], dim=1)) / (n_neighs + 1e-6)
                    cos_s = torch.sum(torch.mean(cos_s * same_c * self.knn[idx, :][:, idx], dim=1)) / (n_neighs + 1e-6)

                # loss
                mse_u, mse_s = mse(u + v_u, d_u), mse(s + v_s, d_s)
                cos_xu, cos_xs = torch.mean(cos_sim(v_u, d_u - u)), torch.mean(cos_sim(v_s, d_s - s))

                mse_u_.append(mse_u.clone().detach().numpy())
                mse_s_.append(mse_s.clone().detach().numpy())
                cos_u_.append(cos_u.clone().detach().numpy())
                cos_s_.append(cos_s.clone().detach().numpy())
                cos_xu_.append(cos_xu.clone().detach().numpy())
                cos_xs_.append(cos_xs.clone().detach().numpy())

                if loss_fun == "mse":
                    if not only_s:
                        if constrain:
                            loss = mse_u + alpha_1 * mse_s - alpha_2 * cos_u - alpha_2 * cos_s
                        else:
                            loss = mse_u + alpha_1 * mse_s
                        loss.backward()
                        optimizer.step()

                    else:
                        if constrain:
                            loss = mse_s - alpha_2 * cos_s
                        else:
                            loss = mse_s
                        loss.backward()
                        optimizer.step()

                elif loss_fun == "cos":
                    if not only_s:
                        if constrain:
                            loss = -cos_xu - alpha_1 * cos_xs - alpha_2 * cos_u - alpha_2 * cos_s
                        else:
                            loss = -cos_xu - alpha_1 * cos_xs
                        loss.backward()
                        optimizer.step()

                    else:
                        if constrain:
                            loss = -cos_xs - alpha_2 * cos_s
                        else:
                            loss = -cos_xs
                        loss.backward()
                        optimizer.step()

            pbar.set_description("DNN training")
            pbar.set_postfix(mse_u=np.around(np.mean(mse_u_), 3), mse_s=np.around(np.mean(mse_s_), 3),
                             cos_u=np.around(np.mean(cos_u_), 3), cos_s=np.around(np.mean(cos_s_), 3),
                             cos_xu=np.around(np.mean(cos_xu_), 3), cos_xs=np.around(np.mean(cos_xs_), 3))

        # coefficient and velocity
        self.dnn.eval()

        if self.normalize:
            v_u, v_s = self.dnn.forward(self.x_u_norm, self.x_s_norm)
        else:
            v_u, v_s = self.dnn.forward(self.x_u, self.x_s)
                
        alpha, beta, gamma = self.dnn.alpha, self.dnn.beta, self.dnn.gamma
        v_u, v_s = alpha - beta * self.x_u, beta * self.x_u - gamma * self.x_s

        alpha, beta, gamma = alpha.clone().detach().numpy(), beta.clone().detach().numpy(), gamma.clone().detach().numpy()
        v_u, v_s = v_u.clone().detach().numpy(), v_s.clone().detach().numpy()
            
        return v_u, v_s, alpha, beta, gamma
    
    
def get_velocity_rate(adata, d_nn, knn, same_c=None, n_dims=[128, 64], n_epochs=100, method="simple", loss_fun="mse",
                      only_s=False, normalize=True, constrain=True, pretrain=False, lr=1e-3, alpha_1=1, alpha_2=0.1,
                      batch_size=512, save_coeff=True, filter_genes=False):

    if filter_genes:
        velocity_genes = compute_velocity_genes(adata, n_top_genes=2000, inplace=False)
        adata = adata[:, velocity_genes]

    x_u = adata.layers["Mu"]
    x_s = adata.layers["Ms"]


    # DNN implementation
    dnn = DNN(x_u, x_s, d_nn, knn, n_dims=n_dims, normalize=normalize)
    if pretrain:
        dnn.pretrain(n_epochs=n_epochs, batch_size=batch_size, loss_fun="cos", only_s=only_s)
    v_u, v_s, alpha, beta, gamma = dnn.fit(n_epochs=n_epochs, batch_size=batch_size, alpha_1=alpha_1,
                                            alpha_2=alpha_2, lr=lr, constrain=constrain, loss_fun="mse",
                                            only_s=only_s, same_c=same_c)
    if save_coeff:
        adata.layers["alpha"] = alpha
        adata.layers["beta"] = beta
        adata.layers["gamma"] = gamma

    adata.layers["velocity"] = v_s
    adata.layers["velocity_u"] = v_u

    return adata, v_u, v_s




import numpy as np
import scanpy as sc
import matplotlib.pyplot as plt
from .start_node import get_start_node
from .prune import graph_prune
from .main_path import get_path, visualize_path


def process_path(adata, group_key, emb_key, njobs=-1, start_mode="stochastic"):
    '''
    :param adata: anndata to use
    :param group_key: cell cluster key to use
    :param emb_key: basis/embedding key to use
    :param filter_subgraph: whether to filter subgraph
    :param njobs: number of parallel jobs
    :return: dictionary of selected main path including start node and following nodes
    '''

    select_info, select_weight = graph_prune(adata, group_key, emb_key)
    start_node = get_start_node(adata, select_info, group_key, njobs=njobs, mode=start_mode)
    start_index = list(select_info.index).index(start_node)

    path_dict = get_path(select_info, select_weight, start_index=start_index)
    ax = visualize_path(select_info, select_weight, path_dict, emb_key, show=False, ax=None)

    adata.uns["path_dict"] = path_dict.copy()
    print("'path_dict' added to adata.uns")

    return path_dict, ax


if __name__ == '__main__':
    # test
    data_path = r"D:\cuhk\project\velocity\dataset\scRNA-seq\BoneMarrow\human_cd34_bone_marrow_processed.h5ad"
    data_name = "bonemarrow"
    group_key = "clusters"
    emb_key = "X_tsne"

    adata = sc.read_h5ad(data_path)
    path_dict, ax = process_path(adata, group_key, emb_key, njobs=-1, start_mode="stochastic")
    print(path_dict)
    print(adata.uns["path_dict"])
    plt.show()










